
/*=============================================================================
   beest, an HTML viewer.
   Copyright (C) 2000 Zachary Thayer Smith

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   The author is reachable by electronic mail at zeyr@excite.com.
=============================================================================*/

/*----------------------------------------------------------------------
 * Module name:    image.c
 * Author name:    Zach Smith
 * Create date:    1 Oct 2000
 * Last change:    7 Dec 2000
 * Purpose:        Image class;
 *                 methods for manipulation of Images;
 *                 methods for reading/writing XBM image files.
 *----------------------------------------------------------------------
 * Note: This module has no association with any color handling routine,
 * so actually, a pixel can contain anything at all so long as it fits 
 * in 8 bits.
 *----------------------------------------------------------------------
 * Changes:
 * 3 Dec 2000, zeyr@excite.com: finished drawing routines except circles
 * 7 Dec 2000, zeyr@excite.com: bugfix in line_draw
 *--------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>


#include "defs.h"
#include "error.h"
#include "main.h"
#include "display.h"
#include "image.h"
#include "malloc.h"
#include "string.h"


Image*
image_new ()
{
	Image *im = (Image*) my_malloc (sizeof (Image));
	if (!im) error_handler ("cannot allocate Image object");

	memset ((void*)im, 0, sizeof(Image));

	return im;
}


void
image_free (Image* im)
{
	CHECK_PARAM_NOT_NULL(im);

	if (im->pixmap) {
		my_free (im->pixmap);
	}
	my_free((char*) im);
}


void
image_write_xbm (Image *im, FILE *f)
{
	Coordinate i,j;
	Dimension w,h;

	CHECK_PARAM_NOT_NULL(f);
	CHECK_PARAM_NOT_NULL(im);

	w=im->width;
	h=im->height;

	fprintf(f,"#define image_width %lu\n", w);
	fprintf(f,"#define image_height %lu\n", h);
	fprintf(f,"static char image_bits[] = {\n");

	for (j=0; j<h; j++) {
		unsigned char *ptr = &im->pixmap[j*w];
		i=0;
		while (i<w) {
			unsigned char ch=0;
			unsigned short mask=1;
			while (i<w && mask<256) {
				if (*ptr++) ch |= mask;
				mask <<= 1;
				i++;
			}
			fprintf (f,"0x%02x, ",ch);
		}
		fprintf (f,"\n");
	}

	fprintf(f,"\n};\n\n");
}




void
image_clear (Image *im,
             unsigned char pixel)
{
	CHECK_PARAM_NOT_NULL(im);

	if (im->pixmap) {
		unsigned long length = im->width * im->height;
		memset ((char*) im->pixmap, pixel, length);
	}
}



/* Routine to copy an image into another image */
void
image_copy (Image *dest, 
	Coordinate dest_x, Coordinate dest_y, 
	Coordinate src_x, Coordinate src_y, 
	Dimension copy_w, Dimension copy_h, Image *src)
{
	Dimension src_w,src_h;
	Dimension dest_w, dest_h;
	Coordinate i,j;
	Coordinate k,l;

	CHECK_PARAM_NOT_NULL(dest);
	CHECK_PARAM_NOT_NULL(src);

	src_w = src->width;
	src_h = src->height;
	dest_w = dest->width;
	dest_h = dest->height;

	if (!src_w || !src_h) 
	  error_handler ("source Image has zero width or height in image_copy");
	if (!dest_w || !dest_h) 
	  error_handler ("destination Image has zero width or height in image_copy");

	for (i=0; i<copy_w; i++) {
		k = i + dest_x;
		if (k < dest_w) {
			for (j=0; j<copy_h; j++) {
				l = j + dest_y;
				if (l < dest_h)
				dest->pixmap [k + l*dest_w] = 
					src->pixmap [src_x+i+(j+src_y)*src_w];
			}
		}
	}
}

	

/* Routine to initialize an image so that it
 * can be drawn into.
 */
void
image_init (Image *im, Dimension width, Dimension height)
{
	unsigned int unit_size = sizeof(unsigned char); /* XX */

	CHECK_PARAM_NOT_NULL(im);

	if (!width || !height) {
		error_handler ("Image has zero width or height in init_image");
	}

	if (debug_mode && logfile) {
		fprintf(logfile, "IMAGE_INIT: created image pixmap %ld by %ld\n",
			width, height);
	}

	im->x = im->y = 0;
	im->width = width;
	im->height = height;
	im->pixmap = my_malloc (width * height * unit_size);
	if (!im->pixmap)
		error_handler ("cannot allocate pixmap for Image in image_init");
}


void
image_copy_from_xbm (Image *dest, Coordinate x, Coordinate y,
	Dimension w, Dimension h, unsigned char *xbm, unsigned long pixel)
{
	Coordinate i,j;
	Dimension dest_w, dest_h;
	unsigned long ww;
	unsigned char *ptr = xbm;
	unsigned char byte;

	CHECK_PARAM_NOT_NULL(dest);
	CHECK_PARAM_NOT_NULL(xbm);

	if (!w || !h) {
		error_handler ("zero width or height Image in copy_xbm_to_image");
	}

	if (w > 1000 || h > 3000) {
		warning_handler ("very large or wide Image in copy_xbm_to_image");
	}

	if (x<0 || y<0) {
		error_handler ("negative x or y in copy_xbm_to_image");
	}

	ww = (w-1) / 8 + 1;
	dest_w = dest->width;	
	dest_h = dest->height;

	/* XBM too tall to fit in Image? */
	if ((h+y)>=dest_h)
		h = dest_h - y;

	for (j=0; j<h; j++) {
		Coordinate k = x;
		unsigned long l = (j+y)*dest_w;
		unsigned char mask=1;
		byte = *ptr++;
		i=0;
		while (i<w) {
			if ((byte & mask) && k<dest_w) {
				dest->pixmap [k+l] =  pixel;
			}
			i++; k++; 
			mask <<= 1;
			if (!mask && i<w) {
				byte = *ptr++;
				mask = 1;
			}
		}
	}
}


void
image_dump (FILE *f, Image *im)
{
	Coordinate i,j;
	Dimension w,h;

	CHECK_PARAM_NOT_NULL(im);

	w=im->width;
	h=im->height;

	fprintf (f, "IMAGE DUMP (%lu x %lu):\n", w,h);
	if (!im->pixmap) {
		fprintf (f, "(no pixmap)\n");
		return;
	}

	for (j=0; j<h; j++) {
		for (i=0; i<w; i++) {
			if (im->pixmap[j*w+i])
				fprintf (f, "#");
			else
				fprintf (f, " ");
		}
		fprintf (f,".\n");
	}
	fprintf (f, "END\n");
}


static int
read_expr (FILE *f, char *str, int len)
{
	int i=0,ch;
	while (EOF != (ch=fgetc(f))) {
		if (ch > ' ') break;
	}
	if (ch==EOF) return FALSE;
	do {
		if (ch <= ' ' || ch == ',' || ch=='}') break;
		*str++ = ch;
		i++;
		if (i==(len-1)) break;
		ch=fgetc(f);
	} while(ch!=EOF);
	*str++=0;
	if (ch==EOF) return FALSE;
	return TRUE;
}


int
image_read_xbm (Image *im, FILE *f, 
	unsigned long set_value, 
	unsigned long reset_value)
{
	Coordinate i,j;
	Dimension w,h;
	unsigned short w2;
	char line[255];
	char *str;
	char *str2;
	unsigned char *ptr;

	do {
		str = fgets (line, 254, f);
		if (!str) return FALSE;
	}
	while (!strchr(str,'#'));

	str2 = strstr (str, "width ");
	if (!str2) return FALSE;
	w = atoi (str2+6);

	do {
		str = fgets (line, 254, f);
		if (!str) return FALSE;
	}
	while (!strchr(str,'#'));

	str2 = strstr (str, "height ");
	if (!str2) return FALSE;
	h = atoi (str2+7);

	do {
		str = fgets (line, 254, f);
		if (!str) return FALSE;
	}
	while (!strstr(str,"static"));

	im->x = im->y = 0;
	im->width = w;
	im->height = h;

	ptr = im->pixmap = (unsigned char*) my_malloc (w*h);
	if (!im->pixmap)
		error_handler ("cannot allocate pixmap for Image while reading xbm file");

	w2 = (w-1)/8+1;

	for (j=0; j<h; j++) {
		Dimension i2=0;
		for (i=0; i<w2; i++) {
			char expr[40];
			int value;
			unsigned char k, mask;

			/* Get eight pixels */
			if (!read_expr (f, expr, 39))
				return FALSE;

			/* Convert it */
			if (*expr=='0' && *(expr+1)=='x') 
				value = h2toi (&expr[2]);
			else 
				value = atoi (expr);
			
			/* Put them into the Image */
			k=0;
			mask=1;
			while (i2<w && k<8) {
				if (value & mask)
					*ptr++ = set_value;
				else 
					*ptr++ = reset_value;

				mask <<= 1;
				k++;
				i2++;
			}
		}
	}

	return TRUE;
}



/*------------Image Graphics Routines-----------------------------------*/


/*====================================================================
 * Vector based graphics for HTML! Too many years have passed without
 * it being implemented. No point in blaming, I'll just do it myself.
 *====================================================================
 */



/* Easy routines first.
 */

void
image_draw_pixel (Image *im, Coordinate x, Coordinate y, unsigned char pixel)
{
	Dimension w,h;

	CHECK_PARAM_NOT_NULL(im);

	/* Can we avoid doing this work? */
	if (!im->pixmap) return;
	if (x<0 || y<0) return;
	w=im->width;
	h=im->height;
	if (x>=w || y>=h) return;

	im->pixmap[y*w+x] = pixel;
}


int
image_get_pixel (Image *im, Coordinate x, Coordinate y)
{
	Dimension w,h;

	CHECK_PARAM_NOT_NULL(im);

	/* Can we avoid doing this work? */
	if (!im->pixmap) return -1;
	if (x<0 || y<0) return -1;
	w=im->width;
	h=im->height;
	if (x>=w || y>=y) return -1;

	return im->pixmap[y*w+x];
}


void
image_draw_hline (Image *im, Coordinate x0, Coordinate x1, Coordinate y, 
                  unsigned char pixel)
{
	Dimension w,h;
	unsigned char*ptr;

	CHECK_PARAM_NOT_NULL(im);

	/* Can we avoid doing this work? */
	if (!im->pixmap) return;
	w=im->width;
	h=im->height;
	if (x0>x1) {
		Coordinate tmp=x0;
		x0=x1;
		x1=tmp;
	}
	if (y < 0 || y>=h) return;
	if (x0 < 0 && x1 < 0) return;
	if (x0>=w && x1>=w) return;

	/* Limits the coordinates to within the image */
	if (x0<0) x0=0;
	if (x1>=w) x1=w-1;

	ptr = &im->pixmap[w*y+x0];
	while (x0<=x1) {
		*ptr++ = pixel;
		x0++;
	}
}

void
image_draw_filled_rect (Image *im, Coordinate x0, Coordinate y0, Coordinate x1,
	Coordinate y1, unsigned char pixel)
{
	Dimension w,h;

	CHECK_PARAM_NOT_NULL(im);

	/* Can we avoid doing this work? */
	if (!im->pixmap) return;
	w=im->width;
	h=im->height;
	if (x0>x1 || y0>y1) return;
	if ((y0<0 && y1<0) || (x0<0 && x1<0)) return;
	if ((x0>=w && x1>=w) || (y0>=h && y1>=h)) return;

	/* Limits the coordinates to within the image */
	if (x0<0) x0=0;
	if (y0<0) y0=0;
	if (x1>=w) x1=w-1;
	if (y1>=h) y1=h-1;

	/* Doing it this way will incur some latencies,
	 * but the code will take up less space.
	 */
	while(y0<=y1) {
		image_draw_hline (im, x0,x1,y0,pixel);
		y0++;
	}
}


void
image_draw_vline (Image *im, Coordinate x, Coordinate y0, Coordinate y1, unsigned char pixel)
{
	Dimension w,h;
	unsigned char*ptr;

	CHECK_PARAM_NOT_NULL(im);

	/* Can we avoid doing this work? */
	if (!im->pixmap) return;
	w=im->width;
	h=im->height;
	if (y0>y1) {
		Coordinate tmp=y0;
		y0=y1;
		y1=tmp;
	}
	if (x < 0 || x>=w) return;
	if (y0 < 0  && y1 < 0) return;
	if (y0>=h && y1>=h) return;

	/* Limits the coordinates to within the image */
	if (y0<0) y0=0;
	if (y1>=h) y1=h-1;

	ptr = &im->pixmap[w*y0+x];
	while (y0<=y1) {
		*ptr = pixel;
		ptr += w;
		y0++;
	}
}


void
image_draw_rect (Image *im, Coordinate x0, Coordinate y0, Coordinate x1,
	Coordinate y1, unsigned char pixel)
{
	CHECK_PARAM_NOT_NULL(im);

	if (!im->pixmap) return;
	if (x0>x1 || y0>y1) return;

	image_draw_hline (im, x0,x1,y0, pixel);
	image_draw_hline (im, x0,x1,y1, pixel);
	image_draw_vline (im, x0,y0,y1, pixel);
	image_draw_vline (im, x1,y0,y1, pixel);
}



/* The well-known algorithm invented by Jack Bresenham.
 * There are faster routines, but I'm working with limited time.
 */
void
image_draw_line (Image *im, Coordinate x0, Coordinate y0, 
                            Coordinate x1, Coordinate y1, 
                            unsigned char pixel)
{
	long dy = y1 - y0;
	long dx = x1 - x0;
	long stepx, stepy;
	Dimension w,h;
	unsigned long index;
	unsigned long max_index;

	CHECK_PARAM_NOT_NULL(im);

	if (debug_mode && logfile) {
		fprintf(logfile, "IMAGE_DRAWLINE: (%ld,%ld)-(%ld,%ld) pixel=%d\n",
			x0,y0,x1,y1,(short)pixel);
	}

	if (x0==x1) {
		image_draw_vline (im,x0,y0,y1,pixel);
		return;
	}

	if (y0==y1) {
		image_draw_hline (im,x0,x1,y0,pixel);
		return;
	}

	w=im->width;
	h=im->height;
	max_index = w*h;

	if (dy < 0) { 
		dy = -dy;
		stepy = -1;
	} else { 
		stepy = 1;
	}
	if (dx < 0) { 
		dx = -dx;
		stepx = -1; 
	} else { 
		stepx = 1; 
	}

	dy <<= 1;
	dx <<= 1;

	// y0 *= w;
	// y1 *= w;

	if (x0>= 0 && y0 >= 0 && x0 < w && y0 < h) {
	// index >=0 && index < max_index)
		index = x0+y0*w;
		im->pixmap [index] = pixel;
	}

	if (dx > dy) {
		int fract = dy - (dx >> 1);
		while (x0 != x1) {
			if (fract >= 0) {
				y0 += stepy;
				fract -= dx;
			}
			x0 += stepx;
			fract += dy;

			if (x0>= 0 && y0 >= 0 && x0 < w && y0 < h) {
			// if (x0>= 0 && y0 >= 0 && x0 < w && y0 < h && index >=0 && index < max_index)
				index = x0+y0*w;
				im->pixmap [index] = pixel;
			}
		}
	} else {
		int fract = dx - (dy >> 1);
		while (y0 != y1) {
			if (fract >= 0) {
				x0 += stepx;
				fract -= dy;
			}
			y0 += stepy;
			fract += dx;

			if (x0>= 0 && y0 >= 0 && x0 < w && y0 < h) {
			// if (x0>= 0 && y0 >= 0 && x0 < w && y0 < h && index >=0 && index < max_index)
				index = x0+y0*w;
				im->pixmap [index] = pixel;
			}
		}
	}
}



/* Somewhat slow routine to draw one character of 
 * a BDF font into an Image. Virtually a carbon copy
 * of my svga_draw_char routine.
 */
void 
image_draw_char (Image *im, Font *fontdata,
		 int simulate_boldface,
		 int simulate_italic,
                 int ch, Coordinate x0, Coordinate y0,
		 unsigned char pixel)
{
	unsigned char *ptn_ptr;
	unsigned char width;
	unsigned short index;
	unsigned char height;
	unsigned char ascent;
	unsigned char bytes_per_line;
	unsigned char bytes_to_draw;
	Coordinate max_x, max_y;
	signed char x_offset;
	signed char y_offset;
	Coordinate x;
	Coordinate y;
	FontSize size;
	int italic_y;

	CHECK_PARAM_NOT_NULL(im);
	CHECK_PARAM_NOT_NULL(fontdata);

	if (!im->pixmap) return;
	if (ch < 32) return;

	size = fontdata->size;
	width = fontdata->widths[ch];
	ascent = fontdata->ascent;
	x_offset = 0; /* fontdata->x_offsets[ch]; */
	y_offset = fontdata->y_offsets[ch];
	index = fontdata->indices[ch];
	bytes_to_draw = fontdata->indices[ch+1]-index;
	ptn_ptr = &fontdata->bytes[index];
	bytes_per_line = ((width-1)/8)+1;
	if (simulate_italic)
		italic_y = ascent;

	max_x = im->width-1;
	max_y = im->height-1;

	if (bytes_per_line < 1 || bytes_per_line > 4) {
		fprintf (logfile, "image_draw_char charno %u: invalid bytes per line in char draw: %u\n",
			(unsigned short) ch, (unsigned short) bytes_per_line); 

		error_handler ("invalid bytes per line in char draw"); 
	}

	height = bytes_to_draw / bytes_per_line;
	x = x0 /* + x_offset */;
	y = y0 + ascent - height - y_offset;

	while (bytes_to_draw > 0) 
	{
		int j, k;
		int offset;

		if (simulate_italic) {
			offset = italic_y-- / 3;
		} else 
			offset=0;

		/* Assemble the pattern */
		k=bytes_per_line;
		j=0;
		while (k--)
		{
			unsigned char byte;
			int n=0;

			byte = *ptn_ptr++;

			while (n!=8 && j<width) {
				if (byte & 0x80) {
					int _x = x+offset;
					if (_x <= max_x &&
					    y <= max_y) {
						im->pixmap[y*im->width+_x] = pixel;
						if (simulate_boldface && (_x+1)<=max_x)
						  im->pixmap[y*im->width+_x+1] = pixel;
					}
				}
				/* byte &= 0x7f; */
				byte <<= 1;
				n++;
				x++;
				j++;
				if (!byte) {
					x += (8-n);
					break;
				}
			}
		}

		x = x0 + x_offset;
		y++;
		bytes_to_draw -= bytes_per_line;
	}
}



/* Somewhat slow routine to draw a text string using
 * a BDF font into an Image. Virtually a carbon copy
 * of my svga_draw_text routine.
 */
void
image_draw_text (Image *im, Font *fontdata,
		 int simulate_boldface,
		 int simulate_italic,
                 String str,
		 Coordinate startx, Coordinate starty,
		 unsigned char pixel)
{
	int ax,ay;
	int ch;

	CHECK_PARAM_NOT_NULL(im);
	CHECK_PARAM_NOT_NULL(fontdata);
	CHECK_PARAM_NOT_NULL(str);

	if (!im->pixmap) return;

	if (debug_mode && logfile)
	  fprintf (logfile, "image_draw_text '%s' start: %ld,%ld pixel:%d\n",
		str, startx, starty, pixel); 

	ax=startx;
	ay=starty;

	while((ch = *str)) {
		if (ch==32) {
			float tmp = 1.0 - 0.6815;
			tmp *= fontdata->ascent;
			ax += tmp;
		} else {
			image_draw_char (im, fontdata,
				simulate_boldface,
				simulate_italic,
				ch, ax, ay, pixel);
			ax += fontdata->widths [ch] + 1;
		}

		if (simulate_boldface) ax+= 1; /* XX should be 2 for 18/24pt? */
		str++;
	}
}


