
/*=============================================================================
   beest, an HTML viewer.
   Copyright (C) 2000 Zachary Thayer Smith

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   The author is reachable by electronic mail at zeyr@excite.com.
=============================================================================*/

/*----------------------------------------------------------------------
 * Module name:    svga.c
 * Author name:    Zach Smith
 * Create date:    1 Oct 2000
 * Last change:    7 Dec 2000
 * Purpose:        Display driver for a VGA or SVGA display;
 *                 supports SVGALIB under Linux;
 *                 supports BGI library under DOS with Turbo C.
 *----------------------------------------------------------------------
 * Changes:
 * 28 Nov 2000, zeyr@excite.com: added draw_line method, popup draw method.
 * 7 Dec 2000, zeyr@excite.com: bugfix in put_image
 * 8 Dec 2000, zeyr@excite.com: bugfix in scrollbar move routines
 *--------------------------------------------------------------------*/

#include <stdio.h>
#include <ctype.h>
#include <string.h>

#ifdef TURBO_C
#include <graphics.h>
#include <conio.h>
#else
#include <vga.h>
#include <vgamouse.h>
#include <vgakeyboard.h>
#endif

#include "defs.h"
#include "malloc.h"
#include "main.h"
#include "display.h"
#include "svga.h"
#include "style.h"
#include "error.h"
#include "special.h"
#include "font.h"
#include "fonthash.h"
#include "bdf.h"
#include "image.h"


/* Our mouse pointer.
 */
#include "left_ptr.xbm"
#include "left_msk.xbm"


/* TO DO
 *--------

  Need to add routine to read the entire contents of a X pixmap from a file.
  DONE Need to add routine to draw bitmap.
  Need to add routine to draw pixmap.
  Need to add routine to read only the dimensions of a graphics file.
 */


#define VSCROLL_WIDTH (8)
#define HSCROLL_HEIGHT (8)
#define VSCROLL_BUTTONSIZE (16)
#define HSCROLL_BUTTONSIZE (16)


/* This is effectively a "graphics context", but it's the only one. 
 */
typedef struct {
	/* Current style */
	Style style;
	int fg; /* this was converted from the Style */
	int bg; /* this was converted from the Style */
	int highlight_bg;

	/* Mouse Pointer */
	unsigned char pointer_is_visible;
	unsigned char pointer_is_hidden;
	Coordinate pointer_x, pointer_y;
	Dimension pointer_width, pointer_height;
	Image *pixels_under_pointer;

	/* Mouse Pointer Colors */
	int pointer_fg;
	int pointer_border_color;
}
svga_data;


/* Just one set, assuming we will only offer a view of
 * only one HTML file (it's only a viewer).
 */
static svga_data sdata;
static DisplayDevice *svga_dpy = NULL;
static int svga_is_open = FALSE;


/*-------Low Level Graphics Routines--------------------------------------*/

static unsigned char _current_color=0; /* fg only */
static int _width=640;
static int _height=480;
/* low-level clip */
static int _x0, _y0, _x1, _y1;

static void _setclip (int x0, int y0, int x1, int y1)
{
	if (x0<0) x0=0;
	if (y0<0) y0=0;
	if (x1>=_width) x1=_width-1;
	if (y1>=_height) y1=_height-1;
	_x0=x0;
	_x1=x1;
	_y0=y0;
	_y1=y1;
}

static int 
_getcolor() {
	return _current_color;
}

static void 
_setcolor (unsigned char c) {
#ifndef TURBO_C
	vga_setcolor(c);
#else
	setcolor(c);
#endif
	_current_color=c;
}

static void 
_putpixel (int x, int y) {
	if (x<_x0 || x>_x1) return;
	if (y<_y0 || y>_y1) return;
#ifndef TURBO_C
	vga_drawpixel (x,y);
#else
	putpixel (x,y,_current_color);
#endif
}

static Pixel 
_getpixel (int x, int y) {
	if (x<_x0 || x>_x1) return -1;
	if (y<_y0 || y>_y1) return -1;
#ifndef TURBO_C
	return vga_getpixel (x,y);
#else
	return getpixel (x,y);
#endif
}

/* rectangular area, no border */
static void 
_color_replace_area (int x0, int y0, int x1, int y1,
	int oldpixel, int newpixel) 
{
	int i,j,k;
	static unsigned char buffer[1024];
	unsigned long count;

	if (x0>x1 || y0>y1) return;
	if (x0<_x0) x0=_x0;
	if (x1>_x1) x1=_x1;
	if (y0<_y0) y0=_y0;
	if (y1>_y1) y1=_y1;
#ifndef TURBO_C
	count=x1-x0+1;

	if (count<0) error_handler ("bad params in _color_replace_area");
	if (count>_width) count=_width;

	for (i=y0; i<=y1; i++) {
		unsigned char* ptr;
		vga_getscansegment (buffer,x0,i,count);
		ptr=buffer;
		j=count;
		while(j--) { if(*ptr==oldpixel)*ptr++ = newpixel; else ptr++; }
		vga_drawscansegment (buffer,x0,i,count);
	}
#else
	k=_getcolor();
	_setcolor(newpixel);
	for (i=y0; i<=y1; i++) {
		for (j=x0; j<=x1; j++) {
			if (_getpixel(j,i)==oldpixel) {
				_putpixel(j,i);
			}
		}
	}
	_setcolor(k);
#endif
}

/* rectangular area, no border */
static void 
_fillarea (int x0, int y0, int x1, int y1) {
	int i;
	static unsigned char buffer[1024];
	unsigned long count;

	if (x0>x1 || y0>y1) return;
	if (x0<_x0) x0=_x0;
	if (x1>_x1) x1=_x1;
	if (y0<_y0) y0=_y0;
	if (y1>_y1) y1=_y1;

#ifndef TURBO_C
	count=x1-x0+1;

	if (count<0) error_handler ("bad params in _fillarea");
	if (count>_width) count=_width;

	for (i=0; i<count; i++) 
		buffer[i] = _current_color;

	for (i=y0; i<=y1; i++) {
		vga_drawscansegment (buffer,x0,i,count);
		// vga_drawline (x0,i,x1,i);
	}
#else
	for (i=y0; i<=y1; i++) {
		line (x0,i,x1,i);
	}
#endif
}

static void 
_drawline (int x0, int y0, int x1, int y1) {
	if (x0 < _x0 && x1 < _x0) return;
	if (x0 > _x1 && x1 > _x1) return;
	if (y0 < _y0 && y1 < _y0) return;
	if (y0 > _y1 && y1 > _y1) return;
#if 0
	if (x0 < _x0) x0=_x0;
	if (x0 >= _width) x0=_width-1;
	if (x1 < 0) x1=0;
	if (x1 >= _width) x1=_width-1;
	if (y0 < 0) y0=0;
	if (y0 >= _height) y0=_height-1;
	if (y1 < 0) y1=0;
	if (y1 >= _height) y1=_height-1;
#endif

	/* XX - need to use our own line-draw routine,
	 * extracted from image.c, in order to enforce 
	 * the clip limits.
	 */

#ifndef TURBO_C
	vga_drawline (x0,y0,x1,y1);
#else
	line (x0,y0,x1,y1);
#endif
}


static void 
_drawrect (int x0, int y0, int x1, int y1) {
	if (x0 < _x0 && x1 < _x0) return;
	if (x0 > _x1 && x1 > _x1) return;
	if (y0 < _y0 && y1 < _y0) return;
	if (y0 > _y1 && y1 > _y1) return;
	_drawline (x0,y0,x0,y1);
	_drawline (x1,y0,x1,y1);
	_drawline (x0,y0,x1,y0);
	_drawline (x0,y1,x1,y1);
}


/* Set a palette element. r,g,b are 8-bit  */
static void
_setpalette (int i, int r, int g, int b) {
#ifndef TURBO_C
	vga_setpalette (i,r>>2,g>>2,b>>2);
#else
	/* XX */
#endif
}


/*-------Image Put/Get-----------------------------------------------------*/

void
svga_get_image (Coordinate x, Coordinate y, Dimension w, Dimension h, Image *im)
{
	Coordinate i,j;

	CHECK_PARAM_NOT_NULL(im);

	if (!w || !h) {
		error_handler ("zero width or height Image in get_image");
	}

	if (w > 1000 || h > 3000) {
		warning_handler ("very large or wide Image in get_image");
	}

	im->x = x;
	im->y = y;
	im->width = w;
	im->height = h;

	if (im->pixmap) {
		error_handler ("Image already has pixmap in get_image");
		warning_handler ("Image already has pixmap in get_image");
		my_free (im->pixmap);
	}

	im->pixmap = my_malloc (w * h * sizeof(Pixel));
	if (!im->pixmap)
		error_handler ("unable to allocate Image pixmap");

	for (i=0; i<w; i++) 
		for (j=0; j<h; j++) 
			im->pixmap [j*w+i] = _getpixel (i+x,j+y);
}




static void
svga_put_image (Coordinate x, Coordinate y, Image *im)
{
	Coordinate istart;
	Coordinate j, jstart;
	Dimension w,h;
	short sw,sh;
	unsigned short max_w, max_h;

	CHECK_PARAM_NOT_NULL(im);

	w=im->width;
	h=im->height;

	if (!w || !h) {
		error_handler ("zero width or height Image in put_image");
	}

	if (w > 1000 || h > 3000) {
		warning_handler ("very large or wide Image in put_image");
	}

	if (!im->pixmap) {
		error_handler ("Image has no pixmap in put_image");
	}

	/* If the image will be completely offscreen,
	 * then we can safely ignore it.
	 */
	sw = svga_dpy->actual_width;
	sh = svga_dpy->actual_height;
	if (x>=sw || y>=sh) {
		return;
	}
	if ((x+w-1)<0 || (y+h-1)<0) {
		return;
	}

	if ((w+x) >= sw)
		max_w = sw - x;
	else
		max_w = w;

	if ((h+y) >= sh)
		max_h = sh - y;
	else
		max_h = h;

	istart = 0;
	if (x<0) istart = -x;

	jstart = 0;
	if (y<0) jstart = -y;

#ifdef TURBO_C
	for (i=istart; i<max_w; i++)  {
		for (j=jstart; j<max_h; j++) {
			_setcolor (im->pixmap [j*w+i]);
			_putpixel (i+x,j+y);
		}
	}
#else
	for (j=jstart; j<max_h; j++) {
		unsigned char *ptr = &im->pixmap[j*w+istart];
		vga_drawscansegment(ptr,x+istart,y+j,max_w-istart);
	}
/*
	SLOW VERSION, LOOKS LIKE A "WIPE". USE LATER?
	for (i=0; i<max_w; i++)  {
		for (j=0; j<max_h; j++) {
			_setcolor (im->pixmap [j*w+i]);
			_putpixel (i+x,j+y);
		}
	}
*/
#endif
}


/* Routine to draw an X bitmap
 */
void
svga_put_xbm (Coordinate x, Coordinate y, 
              Dimension w, Dimension h, unsigned char *xbm,
              Pixel value)
{
	Coordinate i,j,k,l;
	Dimension dest_w, dest_h;
	unsigned long ww;
	unsigned char *ptr = xbm;

	CHECK_PARAM_NOT_NULL(xbm);

	if (!w || !h) {
		error_handler ("zero width or height Image in put_image");
	}

	if (w > 1000 || h > 3000) {
		warning_handler ("very large or wide Image in put_image");
	}

	ww = (w-1) / 8 + 1;

	dest_w = svga_dpy->actual_width;
	dest_h = svga_dpy->actual_height;

	_setcolor (value);
#if 0
	for (j=0; j<h; j++) {
		for (i=0; i<w; i++) {
			if ((1<<(i&7)) & *ptr)
				_putpixel (i+x,j+y);
			if ((i&7)==7) ptr++;
		}
		if (w & 7) 
			ptr++;
	}
#endif
	l = y;
	for (j=0; j<h; j++) {
		unsigned char byte;
		unsigned char mask=1;
		k = x;
		byte = *ptr++;
		i=0;
		while (i<w) {
			if ((byte & mask) && k<dest_w) {
				_putpixel (k,l);
			}
			i++; k++; 
			mask <<= 1;
			if (!mask && i<w) {
				byte = *ptr++;
				mask = 1;
			}
		}
		l++;
	}
}



/*-------Mouse Logic------------------------------------------------------*/

static void 
svga_pointer_undraw (void)
{
	if (sdata.pointer_is_visible) {
		svga_put_image (sdata.pointer_x, sdata.pointer_y, 
		  sdata.pixels_under_pointer);

		image_free (sdata.pixels_under_pointer);
		sdata.pixels_under_pointer = NULL;
	}
}


static void 
svga_pointer_draw (Coordinate x, Coordinate y)
{
#if 0
	if (sdata.pointer_is_visible) {
		svga_pointer_undraw ();
	}
#endif

	/* Save pixels that will be under the pointer */
	if (!sdata.pixels_under_pointer)
		sdata.pixels_under_pointer = image_new();

	svga_get_image (x, y, 
		left_ptr_width,
		left_ptr_height,
		sdata.pixels_under_pointer);

	/* Draw the pointer */
	svga_put_xbm (x, y, left_ptr_width, left_ptr_height, left_ptrmsk_bits,
			sdata.pointer_border_color);
	svga_put_xbm (x, y, left_ptr_width, left_ptr_height, left_ptr_bits,
			sdata.pointer_fg);

	sdata.pointer_x = x;
	sdata.pointer_y = y;
}



static void
svga_pointer_enable (void)
{
	if (!sdata.pointer_is_visible)  {
		svga_pointer_draw(sdata.pointer_x, sdata.pointer_y);
		sdata.pointer_is_visible = TRUE;
	}
}

static void
svga_pointer_disable (void)
{
	if (sdata.pointer_is_visible) {
		svga_pointer_undraw();
		sdata.pointer_is_visible = FALSE;
	}
}


static void
svga_pointer_hide (void)
{
	if (sdata.pointer_is_visible && !sdata.pointer_is_hidden) {
		svga_pointer_undraw();
		sdata.pointer_is_hidden = TRUE;
	}
}


static void
svga_pointer_unhide (void)
{
	if (sdata.pointer_is_visible && sdata.pointer_is_hidden) {
		svga_pointer_draw (sdata.pointer_x, sdata.pointer_y);
		sdata.pointer_is_hidden = FALSE;
	}
}



static void
svga_pointer_move (Coordinate x, Coordinate y)
{
	Coordinate x1,y1;
	Coordinate oldx, oldy;
	Coordinate oldx1, oldy1;

	if (!sdata.pointer_is_visible) {
		// svga_pointer_draw (x,y);
		return;
	}

	x1 = x + left_ptr_width - 1;
	y1 = y + left_ptr_height - 1;

	oldx = sdata.pointer_x;
	oldy = sdata.pointer_y;
	oldx1 = oldx + left_ptr_width - 1;
	oldy1 = oldy + left_ptr_height - 1;

	/* If the two rectangular area occupied by the old and new pointers
	 * do not intersect, our job is easy.
	 */
	if (x1 < oldx || x > oldx1 || y1 < oldy || y > oldy1) {
		Image *tmp;

		/* Get the new location's image */
		tmp = image_new();
		svga_get_image (x, y, left_ptr_width, left_ptr_height, tmp);

		/* Draw the new pointer */
		svga_put_xbm (x, y, left_ptr_width, left_ptr_height, 
			left_ptrmsk_bits, 
			sdata.pointer_border_color);
		svga_put_xbm (x, y, left_ptr_width, left_ptr_height, 
			left_ptr_bits,
			sdata.pointer_fg);

		/* Write the old locations original */
		svga_put_image (sdata.pointer_x, sdata.pointer_y, 
		  sdata.pixels_under_pointer);

		/* Copy the image over */
		image_free (sdata.pixels_under_pointer);
		sdata.pixels_under_pointer = tmp;
		sdata.pointer_x = x;
		sdata.pointer_y = y;
	} else {
		Image *tmp;
		Image *tmp2;
		Coordinate x0,y0;
		Coordinate x2,y2;

		/* Determine the size of an Image which will enclose
	 	 * both rectangles.
	 	 */
		x0 = x<oldx ? x : oldx;
		y0 = y<oldy ? y : oldy;
		x2 = x1>oldx1 ? x1 : oldx1;
		y2 = y1>oldy1 ? y1 : oldy1;

		/* Get the enclosing area. */
		tmp = image_new();
		tmp2 = image_new();
		svga_get_image (x0, y0, x2-x0+1, y2-y0+1, tmp);

		/* Overwrite the old pointer's area into the new Image */
		image_copy (tmp, oldx-x0, oldy-y0,
			0, 0,
			sdata.pixels_under_pointer->width,
			sdata.pixels_under_pointer->height,
			sdata.pixels_under_pointer);

		/* Get the new pointer's area. */
		image_init (tmp2, left_ptr_width, left_ptr_height);
		image_copy (tmp2, 0, 0,
			x-x0, y-y0, 
			left_ptr_width, left_ptr_height,
			tmp);

		/* Draw the xbm into the bigger area */
		image_copy_from_xbm (tmp, x-x0, y-y0, left_ptr_width,
			left_ptr_height, left_ptrmsk_bits, 
			sdata.pointer_border_color);
		image_copy_from_xbm (tmp, x-x0, y-y0, left_ptr_width,
			left_ptr_height, left_ptr_bits, 
			sdata.pointer_fg);

		/* Copy the new image onto the screen */
		svga_put_image (x0, y0, tmp);

		/* Clean up */
		image_free (tmp);
		sdata.pixels_under_pointer = tmp2;
		sdata.pointer_x = x;
		sdata.pointer_y = y;
	}
	sdata.pointer_is_visible = TRUE;
}

/*-------Color Translation------------------------------------------------*/


static Color color_table [256];
static unsigned char color_table_entry_in_use [256];
static String color_names [256];
static int total_colors=0;
static int max_colors=256;


static int
color_table_add (String name)
{
	int color_num;
	Color c;

	if (total_colors >= max_colors) {
		/* XX need to look for removed colors */
		return -1;
	}

	c = color_lookup (name);
	if (c==INVALID_COLOR)
		error_handler ("cannot translate color");

	color_num = total_colors;
	color_table[color_num] = c;
	color_names[color_num] = name; /* NO strdup! */
	color_table_entry_in_use [color_num] = TRUE;
	total_colors++;

	_setpalette (color_num, (c>>16)&0xff, (c>>8)&0xff, c&0xff);

	if (debug_mode && logfile) {
		fprintf(logfile, "COLOR_TABLE_ADD: color %d is now %s (0x%08lx)\n",
			color_num, name, c);
		fflush(logfile);
	}

	return color_num;
}


static int
color_table_lookup (String name)
{
	int i;
	i=0;
	while (i<total_colors) {
		if (color_table_entry_in_use [i] &&
		    my_case_insensitive_strcmp (name, color_names[i]))
		{
			return i;
		}
		i++;
	}

	/* Always print a failed color lookup to the logfile.
	 */
	if (logfile) {
		fprintf (logfile, "Color lookup failed: '%s'\n", name);
		fflush(logfile);
	}
	return -1;
}


static int
color_table_lookup_by_rgb (Color color)
{
	int i;

#if 0
	int r,g,b;

	USEFUL.

	r = (fg>>16) & 0xff;
	g = (fg>>8) & 0xff;
	b = fg & 0xff;

	r = (bg>>16) & 0xff;
	g = (bg>>8) & 0xff;
	b = bg & 0xff;
#endif

	i=0;
	while (i<total_colors) {
		if (color_table_entry_in_use [i] &&
		    color == color_table[i])
		{
			return i;
		}

		/* XX - need to determine the closest color in case
		 * there's no exact match.
		 */

		i++;
	}

	/* Always print a failed color lookup to the logfile.
	 */
	if (logfile) {
		fprintf (logfile, "Color lookup failed: 0x%08lx\n", color);
		fflush(logfile);
	}
	return -1;
}


static int
color_table_remove (int color_num)
{
	if (color_num<0 || color_num>=max_colors) return FALSE;

	color_table_entry_in_use[color_num] = FALSE;
	return TRUE;
}



static int
color_table_init (int max)
{
	int color;
	int i;

	max_colors = max;

	for (i=0; i<max; i++) 
		color_table_entry_in_use[i]=FALSE;

	total_colors=0;

	color = color_table_add ("black");
	if (color<0) return FALSE;
	svga_dpy->black = color;

	color = color_table_add ("red");
	if (color<0) return FALSE;
	svga_dpy->red = color;

	color = color_table_add ("green");
	if (color<0) return FALSE;
	svga_dpy->green = color;

	color = color_table_add ("blue");
	if (color<0) return FALSE;
	svga_dpy->blue = color;

	color = color_table_add ("cyan");
	if (color<0) return FALSE;

	color = color_table_add ("yellow");
	if (color<0) return FALSE;

	color = color_table_add ("magenta");
	if (color<0) return FALSE;

	color = color_table_add ("white");
	if (color<0) return FALSE;
	svga_dpy->white = color;

	/* Now assuming we are using VGA16 mode, let's set up 
	 * eight greys to allow us to nicely render color images 
	 * without the bother of dithering, although of course 
	 * they will be greyscale, not color.
	 */
	color = color_table_add ("grey1");
	if (color<0) return FALSE;
	svga_dpy->grey1 = color;

	color = color_table_add ("grey2");
	if (color<0) return FALSE;
	svga_dpy->grey2 = color;

	color = color_table_add ("grey3");
	if (color<0) return FALSE;
	svga_dpy->grey3 = color;

	color = color_table_add ("grey4");
	if (color<0) return FALSE;
	svga_dpy->grey4 = color;

	color = color_table_add ("grey5");
	if (color<0) return FALSE;
	svga_dpy->grey5 = color;

	color = color_table_add ("grey6");
	if (color<0) return FALSE;
	svga_dpy->grey6 = color;

	color = color_table_add ("grey7");
	if (color<0) return FALSE;
	svga_dpy->grey7 = color;

	color = color_table_add ("grey8");
	if (color<0) return FALSE;
	svga_dpy->grey8 = color;

	return TRUE;
}



/*-------SVGA Methods---------------------------------------------------*/

void
svga_clear () {
	int color=_getcolor();

	/* XX if we add a title bar, this will change: */
	_setcolor(sdata.bg);
	_fillarea (0,0,svga_dpy->width-1,svga_dpy->height-1);
	_setcolor (color);
}

static void
svga_set_colors (Color fg, Color bg)
{
	if (!svga_is_open)
	  error_handler ("attempt to set svga colors before svga is opened");

	sdata.fg = color_table_lookup_by_rgb (fg);
	sdata.bg = color_table_lookup_by_rgb (bg);

	if (sdata.fg < 0 || sdata.bg < 0)
		error_handler ("unable to assign color");

	_setcolor (sdata.fg);
}



static void
svga_reset_colors ()
{
	if (!svga_is_open)
	  error_handler ("attempt to set svga colors before svga is opened");

	sdata.fg = color_table_lookup_by_rgb (svga_dpy->fgcolor);
	sdata.bg = color_table_lookup_by_rgb (svga_dpy->bgcolor);

	if (sdata.fg < 0 || sdata.bg < 0)
		error_handler ("unable to assign default color");

	_setcolor (sdata.fg);
}

int
svga_special_expr (String str)
{
	int ch=0x80;
	char tmp[20];
	String tmp2;

	/* Remove the & and ; for comparison */
	if ('&'==*str)
		strcpy (tmp,str+1);
	else
		strcpy (tmp,str);
	tmp2 = strchr(tmp,';');
	if(tmp2) *tmp2=0;

	ch = special_translate (tmp);

	return ch;
}


/* Routine to provide dimensions of a string.
 * If the current Font is invalid e.g. if the font family is unknown or
 * unspecified, then it is HERE (not before, not in the tag processing)
 * that we switch to the default).
 */

void
svga_text_dims (String str, Style *sl, 
	Dimension *w, 
	Dimension *ascent,
	Dimension *descent) 
{
	Dimension total_width = 0;
	unsigned char simulate_boldface;
	int ch;

	CHECK_PARAM_NOT_NULL(str);
	CHECK_PARAM_NOT_NULL(sl);
	CHECK_PARAM_NOT_NULL(w);
	CHECK_PARAM_NOT_NULL(ascent);
	CHECK_PARAM_NOT_NULL(descent);

	if (!svga_is_open)
	  error_handler ("attempt to get text dimensions before svga is opened");

	simulate_boldface = sl->flags & ATTR_BOLD;

	style_fix_font (svga_dpy, sl);

	while ((ch=*str)) 
	{
		if (ch>=32)
		{
			{
				total_width += sl->fontdata->widths[ch]+1;
				if (ch==32)
				    total_width += 
					sl->fontdata->widths [ch] + 1;

				if (simulate_boldface) {
					total_width += 1;
				}
			}
		}

		str++;
	}

	*w = total_width;

	if (sl->flags & ATTR_OVERBAR)
		*ascent = sl->fontdata->ascent + 1;
	else
		*ascent = sl->fontdata->ascent;

	if (sl->flags & ATTR_UNDERLINE)
		*descent = sl->fontdata->descent + 2;
	else
		*descent = sl->fontdata->descent;
}


void
svga_draw_x (Coordinate x0, Coordinate y0, Coordinate x1, Coordinate y1)
{
	if (!svga_is_open)
	  error_handler ("attempt to draw rectangle on svga before svga is opened");

	if (x0>x1 || y0>y1) return;

	_drawline (x0,y0,x1,y1);
	_drawline (x0,y1,x1,y0);
}


/* draw a horizontal line */
void svga_draw_hline (Coordinate x0, Coordinate x1, Coordinate y) {
	Coordinate w;

	if (!svga_is_open)
	  error_handler ("attempt to draw line on svga before svga is opened");

	w = svga_dpy->width;

	if (x1 < 0 || x0 >= w) return;
	if (y < 0 || y>= svga_dpy->height) return;

	if (x0 < 0) x0 = 0;
	if (x1 >= w) x1 = w-1;

	_setcolor(sdata.fg);
	_drawline (x0,y,x1,y);
}

/* draw a vertical line */
void svga_draw_vline (Coordinate x, Coordinate y0, Coordinate y1) {
	Coordinate h;

	if (!svga_is_open)
	  error_handler ("attempt to draw line on svga before svga is opened");

	h = svga_dpy->height;

	if (y1 < 0 || y0 >= h) return;
	if (x < 0 || x>= svga_dpy->width) return;

	if (y0 < 0) y0 = 0;
	if (y1 >= h) y1 = h-1;

	_setcolor(sdata.fg);
	_drawline (x,y0,x,y1);
}

/* draw a line */
void svga_draw_line (Coordinate x0, Coordinate y0, Coordinate x1, Coordinate y1) {
	if (!svga_is_open)
	  error_handler ("attempt to draw line on svga before svga is opened");

	_setcolor(sdata.fg);
	_drawline (x0,y0,x1,y1);
}



static void
svga_draw_text (Coordinate startx, Coordinate starty, String str,
                unsigned char has_following_whitespace) 
{
	/* per char */
	unsigned char *ptn_ptr;
	unsigned char width;
	unsigned short index;
	unsigned char height;
	unsigned char ascent;
	unsigned char bytes_per_line;
	unsigned char bytes_to_draw;
	short max_x, max_y;
	signed char x_offset;
	signed char y_offset;
	Font *fontdata;
	short x;
	short y;
	unsigned char our_flags=0;
	int italic_y;

	/* per text string */
	int ax;
	int ch;

	CHECK_PARAM_NOT_NULL(str);

	if (!svga_is_open)
	  error_handler ("attempt to draw text on svga before svga is opened");

	if (debug_mode && logfile) {
	  	fprintf (logfile, "SVGA draw_text '%s' start: %ld,%ld\n",
			str, startx, starty); 
		fflush(logfile);
	}

	style_fix_font (svga_dpy, &sdata.style);

	if (!sdata.style.fontdata)
	  error_handler ("text draw cannot find valid Font object");

	max_x = svga_dpy->actual_width - svga_dpy->vscroll_width;
	max_y = svga_dpy->actual_height - svga_dpy->hscroll_height;

	our_flags |= sdata.style.flags & ATTR_BOLD ? 1 : 0;
	our_flags |= sdata.style.flags & ATTR_ITALIC ? 2 : 0;
	our_flags |= sdata.style.flags & ATTR_OVERBAR ? 4 : 0;
	fontdata = sdata.style.fontdata;
	ascent = fontdata->ascent;

	ax=startx;

	_setcolor(sdata.fg);

	while((ch = *str)) {
		short y0 = starty;

		width = fontdata->widths[ch];
		x_offset = 0; /* fontdata->x_offsets[ch]; */
		y_offset = fontdata->y_offsets[ch];
		index = fontdata->indices[ch];
		bytes_to_draw = fontdata->indices[ch+1]-index;
		ptn_ptr = &fontdata->bytes[index];
		bytes_per_line = ((width-1)/8)+1;
		if (our_flags & 2)
			italic_y = ascent;

		if (bytes_per_line < 1 || bytes_per_line > 4) {
			fprintf (logfile, "charno %u: invalid bytes per line in char draw: %u\n",
				(unsigned short) ch, (unsigned short) bytes_per_line); 
			fflush(logfile);

			/* style_dump (logfile, &sdata.style); */

			error_handler ("invalid bytes per line in char draw"); 
		}

		height = bytes_to_draw / bytes_per_line;
		x = ax;
		y = y0 + ascent - height - y_offset;

		if (our_flags&4) 
			y += 1;

		while (bytes_to_draw > 0) 
		{
			int j, k;
			int offset;

			if (our_flags & 2) {
				offset = italic_y-- / 3;
			} else 
				offset=0;

			/* Assemble the pattern */
			k=bytes_per_line;
			j=0;
			while (k--)
			{
				unsigned char byte;
				int n=0;

				byte = *ptn_ptr++;

				while (n!=8 && j<width) {
					if (byte & 0x80) {
						short _x = x+offset;
						if (_x < max_x &&
						    y < max_y) {
							_putpixel (_x, y);
							if ((our_flags&1) && (_x+1)<max_x)
								_putpixel (_x+1, y);
						}
					}
					/* byte &= 0x7f; */
					byte <<= 1;
					n++;
					x++;
					j++;
					if (!byte) {
						x += (8-n);
						break;
					}
				}
			}

			x = ax /* + x_offset */;
			y++;
			bytes_to_draw -= bytes_per_line;
		}
/*================*/

		ax += sdata.style.fontdata->widths [ch] + 1;
		if (our_flags & 1) ax+= 1; /* XX should be 2 for 18/24pt? */
		str++;
	}

	if (has_following_whitespace)
		ax += sdata.style.fontdata->ascent * (1.0-0.6815);

	_setcolor(svga_dpy->black);

	if (our_flags & 4) {
		svga_draw_hline (startx,ax,starty+1);
	}

	if (sdata.style.flags & ATTR_UNDERLINE) {
		svga_draw_hline (startx,ax,starty+ascent+2);
	}

	if (sdata.style.flags & ATTR_STRIKE) {
		svga_draw_hline (startx,ax,starty+ascent-ascent/3-1);
	}
}


void
svga_draw_twoline_centered_popup (String s1, String s2, Image *im)
{
	Dimension w,a,d;
	Dimension total_width;
	Dimension total_height;
	Coordinate x,y;
	Dimension padding=15;
	Coordinate x1,y1;

	CHECK_PARAM_NOT_NULL(s1);
	CHECK_PARAM_NOT_NULL(s2);
	CHECK_PARAM_NOT_NULL(im);

	svga_text_dims (s1, &sdata.style, &w, &a, &d);
	total_width = w;
	total_height = a+d;
	if (s2 && *s2) {
		svga_text_dims (s2, &sdata.style, &w, &a, &d);
		if (w>total_width) total_width=w;
		total_height += a+d;
	}

	x = svga_dpy->actual_width - total_width - padding*2;
	x /= 2;
	y = svga_dpy->actual_height - total_height - padding*2;
	y /= 2;

	total_width += padding*2;
	total_height += padding*2;

	x1=x+total_width-1;
	y1=y+total_height-1;

	svga_get_image (x,y,total_width,total_height,im);
	if (!im->pixmap)
		error_handler ("get_image did not return a pixmap");

	_setcolor(sdata.bg);
	_fillarea(x,y,x1,y1);
	_setcolor(sdata.fg);
	_drawrect(x,y,x1,y1);
	_drawrect(x+1,y+1,x1-1,y1-1);
	svga_draw_text(x+padding,y+padding,s1,FALSE);
	if (s2 && *s2)
	svga_draw_text(x+padding,
           y+padding+sdata.style.fontdata->ascent+sdata.style.fontdata->descent,
	   s2,FALSE);
}

void 
svga_draw_rectangle (Coordinate x0, Coordinate y0, Coordinate x1, Coordinate y1) 
{
	if (!svga_is_open)
	  error_handler ("attempt to draw rectangle on svga before svga is opened");

fprintf (logfile,"drawing rectangle %d %d - %d %d\n",x0,y0,x1,y1); 

	svga_draw_hline (x0,x1,y0);
	svga_draw_hline (x0,x1,y1);
	svga_draw_vline (x0,y0,y1);
	svga_draw_vline (x1,y0,y1);
}


/* Draw a simple 3d rectangle.
 * Note! This is not intended to be lifelike at all.
 */
void 
svga_draw_3d_rectangle (Coordinate x0, Coordinate y0, 
                        Coordinate x1, Coordinate y1,
                        Dimension size,
                        Color top, Color left, Color bottom, Color right) 
{
	long i;
	int t,b,l,r;

	if (!svga_is_open)
	  error_handler ("attempt to draw 3d rectangle on svga before svga is opened");

	t = color_table_lookup_by_rgb (top);
	b = color_table_lookup_by_rgb (bottom);
	l = color_table_lookup_by_rgb (left);
	r = color_table_lookup_by_rgb (right);

	_setcolor(t);
	_fillarea(x0+size,y0,x1-size+1,y0+size-1);

	_setcolor(b);
	_fillarea(x0+size,y1-size+1,x1-size+1,y1);

	_setcolor(l);
	_fillarea(x0,y0+size,x0+size-1,y1-size+1);

	_setcolor(r);
	_fillarea(x1-size+1,y0+size,x1,y1-size+1);

	/* XX -- speedup: do this with images instead */

	for (i=0; i < size; i++) {
		_setcolor(t);
		_drawline (x0+i,y0+i,x0+size,y0+i);
		_drawline (x1-size,y0+i,x1-i,y0+i);
		_setcolor(b);
		_drawline (x0+i,y1-i,x0+size,y1-i);
		_drawline (x1-size,y1-i,x1-i,y1-i);
		_setcolor(l);
		if (i) _drawline (x0,y0+i,x0+i,y0+i);
		if (i) _drawline (x0,y1-i,x0+i,y1-i);
		_setcolor(r);
		if (i) _drawline (x1,y0+i,x1-i,y0+i);
		if (i) _drawline (x1,y1-i,x1-i,y1-i);
	}
}

void 
svga_filled_rectangle (Coordinate x0, Coordinate y0, Coordinate x1, Coordinate y1) 
{
	if (!svga_is_open)
	  error_handler ("attempt to draw rectangle on svga before svga is opened");

/* fprintf (logfile,"drawing filled rectangle %d %d - %d %d\n",x0,y0,x1,y1); */

	_setcolor(sdata.fg);
	_fillarea (x0,y0,x1,y1);
}


void 
svga_highlight_area (Coordinate x0, Coordinate y0, Coordinate x1, Coordinate y1) 
{
	if (!svga_is_open)
	  error_handler ("attempt to draw rectangle on svga before svga is opened");

/* fprintf (logfile,"drawing filled rectangle %d %d - %d %d\n",x0,y0,x1,y1); */

	_setcolor(sdata.fg);
	_color_replace_area (x0,y0,x1,y1,
		svga_dpy->bg,
		sdata.highlight_bg);
}


void 
svga_unhighlight_area (Coordinate x0, Coordinate y0, Coordinate x1, Coordinate y1) 
{
	if (!svga_is_open)
	  error_handler ("attempt to draw rectangle on svga before svga is opened");

/* fprintf (logfile,"drawing filled rectangle %d %d - %d %d\n",x0,y0,x1,y1); */

	_setcolor(sdata.fg);
	_color_replace_area (x0,y0,x1,y1,
		sdata.highlight_bg,
		svga_dpy->bg);
}


void 
svga_erase_rectangle (Coordinate x0, Coordinate y0, Coordinate x1, Coordinate y1) 
{
	if (!svga_is_open)
	  error_handler ("attempt to draw rectangle on svga before svga is opened");

/* fprintf (logfile,"drawing rectangle %d %d - %d %d\n",x0,y0,x1,y1); */

	_setcolor(sdata.bg);
	_fillarea(x0,y0,x1,y1);
	_setcolor(sdata.fg);
}


void 
svga_draw_shaded_rectangle (Coordinate x0, Coordinate y0, Coordinate x1, Coordinate y1) 
{
	int i;
	int j;

	if (!svga_is_open)
	  error_handler ("attempt to draw rectangle on svga before svga is opened");

/* fprintf (logfile,"drawing rectangle %d %d - %d %d\n",x0,y0,x1,y1); */

	_setcolor(sdata.fg);
	_drawrect(x0,y0,x1,y1);
	_setcolor(sdata.bg);
	_fillarea(x0+1,y0+1,x1-1,y1-1);
	_setcolor(sdata.fg);
	for (i=y0+1; i<y1; i++) {
		j = x0+(i&1);
		while (j<x1) {
			_putpixel (j,i);
			j+=2;
		}
	}
}


void
svga_draw_blank_rectangle (Coordinate x0, Coordinate y0, Coordinate x1, Coordinate y1) 
{
	if (!svga_is_open)
	  error_handler ("attempt to draw rectangle on svga before svga is opened");

/* fprintf (logfile,"drawing rectangle %d %d - %d %d\n",x0,y0,x1,y1); */

	_setcolor(sdata.fg);
	_drawrect(x0,y0,x1,y1);
	_setcolor(sdata.bg);
	_fillarea(x0+1,y0+1,x1-1,y1-1);
}




void
svga_vscroll_draw ()
{
	int start_x;
	int end_y;
	long position;
	unsigned char do_hide;

	if (!svga_dpy->vscroll_width) return;

	svga_reset_colors();

	/* This could be optimized to exclude use
	 * of the rectangle routines.
	 */

	start_x = svga_dpy->actual_width - svga_dpy->vscroll_width;
	end_y = svga_dpy->actual_height - svga_dpy->hscroll_height;

	do_hide = sdata.pointer_x >= start_x;
	if (do_hide)
		svga_pointer_hide();

	svga_draw_shaded_rectangle (start_x, 0, 
		svga_dpy->actual_width-1, end_y);

	position = svga_dpy->vscroll_value * 
              (svga_dpy->actual_height - svga_dpy->hscroll_height - VSCROLL_BUTTONSIZE);
	position /= 100;

	svga_draw_blank_rectangle (start_x, position, 
		svga_dpy->actual_width-1, position+VSCROLL_BUTTONSIZE-1);

	if (do_hide)
		svga_pointer_unhide();
}


void
svga_vscroll_undraw ()
{
	int start_x;
	int end_y;
	int i;
	unsigned char do_hide;

	if (!svga_dpy->vscroll_width) return;

	/* This could be optimized to exclude use
	 * of the rectangle routines.
	 */

	start_x = svga_dpy->actual_width - svga_dpy->vscroll_width;
	end_y = svga_dpy->actual_height - svga_dpy->hscroll_height;

	do_hide = sdata.pointer_x >= start_x;
	if (do_hide)
		svga_pointer_hide();

	i = _getcolor();
	_setcolor(sdata.bg);
	_fillarea (start_x,0, svga_dpy->actual_width-1, end_y);
	_setcolor(i);

	if (do_hide)
		svga_pointer_unhide();
}


void
svga_hscroll_draw ()
{
	int start_y;
	int end_x;
	long position;
	unsigned char do_hide;

	if (!svga_dpy->hscroll_height) return;

	svga_reset_colors();

	/* This could be optimized to exclude use
	 * of the rectangle routines.
	 */

	start_y = svga_dpy->actual_height - svga_dpy->hscroll_height;
	end_x = svga_dpy->actual_width - svga_dpy->vscroll_width - 1;

	do_hide = sdata.pointer_y >= start_y;
	if (do_hide)
		svga_pointer_hide();

	svga_draw_shaded_rectangle (0, start_y, end_x, 
		svga_dpy->actual_height-1);

	position = svga_dpy->vscroll_value * 
	     (svga_dpy->actual_width - svga_dpy->vscroll_width - HSCROLL_BUTTONSIZE);
	position /= 100;

	svga_draw_blank_rectangle (position, start_y, 
		position+HSCROLL_BUTTONSIZE-1, svga_dpy->actual_height-1);

	if (do_hide)
		svga_pointer_unhide();
}


void
svga_hscroll_undraw ()
{
	int start_y;
	int end_x;
	int i;
	unsigned char do_hide;

	if (!svga_dpy->hscroll_height) return;

	/* This could be optimized to exclude use
	 * of the rectangle routines.
	 */

	start_y = svga_dpy->actual_height - svga_dpy->hscroll_height;
	end_x = svga_dpy->actual_width - svga_dpy->vscroll_width - 1;

	do_hide = sdata.pointer_y >= start_y;
	if (do_hide)
		svga_pointer_hide();

	i = _getcolor();
	_setcolor(sdata.bg);
	_fillarea (0, start_y, end_x, 
		svga_dpy->actual_height-1);
	_setcolor(i);

	if (do_hide)
		svga_pointer_unhide();
}


static void
svga_vscroll_enable()
{
	svga_dpy->vscroll_width=VSCROLL_WIDTH;
	svga_dpy->vscroll_value=0;
	svga_dpy->width = svga_dpy->actual_width - VSCROLL_WIDTH;
	svga_vscroll_draw();
}


static void
svga_vscroll_disable()
{
	svga_vscroll_undraw();
	svga_dpy->vscroll_width=0;
	svga_dpy->vscroll_value=0;
	svga_dpy->width = svga_dpy->actual_width;
}


void
svga_hscroll_enable()
{
	svga_dpy->hscroll_height=HSCROLL_HEIGHT;
	svga_dpy->hscroll_value=0;
	svga_dpy->height = svga_dpy->actual_height - HSCROLL_HEIGHT;
	svga_hscroll_draw();
}

void
svga_hscroll_disable()
{
	svga_hscroll_undraw();
	svga_dpy->hscroll_height=0;
	svga_dpy->hscroll_value=0;
	svga_dpy->height = svga_dpy->actual_height;
}


void
svga_vscroll_move(int value)
{
	unsigned char do_hide = 
	 sdata.pointer_x >= (svga_dpy->actual_width - svga_dpy->vscroll_width);
	if (do_hide)
		svga_pointer_hide();

	if (value<0) value=0;
	if (value>100) value=100;
	svga_dpy->vscroll_value=value;
	svga_vscroll_draw();

	if (do_hide)
		svga_pointer_unhide();
}

void
svga_hscroll_move(int value)
{
	unsigned char do_hide = 
	 sdata.pointer_y >= (svga_dpy->actual_height - svga_dpy->hscroll_height);
	if (do_hide)
		svga_pointer_hide();

	if (value<0) value=0;
	if (value>100) value=100;
	svga_dpy->hscroll_value=value;
	svga_hscroll_draw();

	if (do_hide)
		svga_pointer_unhide();
}



static void
svga_allocate_buffers (Dimension width, Dimension height) {
	if (!svga_dpy)
		error_handler ("attempt to allocate display buffers before DisplayDevice data created");

	/* SVGA graphics is not buffered */
}


/* Because we do NO buffering of all displayed data, we need to call the
 * client's resize callback.
 */
void 
svga_resize (Dimension new_width, Dimension new_height) {
	if (!svga_is_open)
	  error_handler ("attempt to resize svga before svga is opened");
	if (!svga_dpy)
	  error_handler ("attempt to resize svga before DisplayDevice data allocated");

	if (svga_dpy->resize_callback) {
	}
}


void
svga_set_clip (Coordinate x0, Coordinate y0, Coordinate x1, Coordinate y1) 
{
	if (!svga_is_open)
	  error_handler ("attempt to set svga clip before svga is opened");

	_setclip (x0,y0,x1,y1);
}


void
svga_reset_clip ()
{
	if (!svga_is_open)
	  error_handler ("attempt to set svga clip before svga is opened");

	_setclip (0,0,svga_dpy->actual_width-1,svga_dpy->actual_height-1);
}


void 
svga_close () {
	if (svga_is_open) {

#ifdef TURBO_C
		closegraph();
#else

		if (debug_mode && logfile) {
			int i;
			int r,g,b;
			fprintf (logfile, "SVGA palette:\n");
			for (i=0; i<max_colors;i++) {
				vga_getpalette(i,&r,&g,&b);
				fprintf (logfile, "color %d: red %d green %d blue %d\n", i,r,g,b);
			}
			fflush(logfile);
		}
		vga_clear();
		mouse_close();
		vga_screenoff();
#endif

		display_free (svga_dpy);
		svga_is_open = FALSE;
	}
}


int
svga_get_event (DisplayEvent *de) {
	int i;
	CHECK_PARAM_NOT_NULL(de);

#ifdef TURBO_C
	0 72 = up
	0 73 = pgup
	0 80 = dn
	0 81 = pgdn
#else
	if (mouse_update()) {
		de->value = mouse_getbutton();
		de->x0 = mouse_getx();
		de->y0 = mouse_gety();
		de->type = EVENT_MOUSE;
		if (!sdata.pointer_is_visible)
			svga_pointer_enable();
		svga_pointer_move (de->x0, de->y0);
		return TRUE;
	}
	i = vga_getkey();
	if (i) {
		if (i==27) {
			i = vga_getkey();
			if (i==0x5b) {
				i = vga_getkey();
				if (i==66) {
					de->type = EVENT_DOWN;
					return TRUE;
				} else if (i==65) {
					de->type = EVENT_UP;
					return TRUE;
				} else if (i==68) {
					de->type = EVENT_LEFT;
					return TRUE;
				} else if (i==67) {
					de->type = EVENT_RIGHT;
					return TRUE;
				} else if (i==53) {
					de->type = EVENT_PGUP;
					vga_getkey(); // skip "~"
					return TRUE;
				} else if (i==54) {
					de->type = EVENT_PGDN;
					vga_getkey(); // skip "~"
					return TRUE;
				} else if (i==49) {
					de->type = EVENT_HOME;
					vga_getkey(); // skip "~"
					return TRUE;
				} else if (i==52) {
					de->type = EVENT_END;
					vga_getkey(); // skip "~"
					return TRUE;
				} else {
fprintf (logfile, "UNKNOWN ESCAPE : %d\n", i);
				}
			}
		}
		else
		{
			de->type = EVENT_KEY;
			de->value = i;
			return TRUE;
		}
	}
#endif

	return FALSE;
}




/* Because we do NO buffering of display data, we must
 * call the client's refresh callback.
 */
static
void svga_refresh (Coordinate rx, Coordinate ry, 
	Coordinate x0, Coordinate y0, Coordinate x1, Coordinate y1)
{
	if (!svga_is_open)
	  error_handler ("attempt to refresh svga before svga is opened");

	_setcolor(sdata.fg);
	_drawline (x0,y0,x1,y1);
	_drawline (x0,y1,x1,y0);
	_setcolor(sdata.bg);

	if (svga_dpy->refresh_callback) {
/* fprintf (logfile, "calling refresh callback\n"); */
#if 0
		svga_dpy->refresh_callback (svga_dpy,rx,ry,x0,y0,x1,y1);
#endif
	} else {
/* fprintf (logfile,"unable to call refresh callback\n"); */
#if 0
		warning_handler ("request for refresh but no callback registered");
#endif
	}
}


/* Diagnostic dump */
static void 
svga_dump ()
{
	if (!svga_is_open)
	  error_handler ("attempt to dump svga before svga is opened");

	/* group_dump() is more appropriate, or a screen dump */
}


static void
svga_set_style (Style *sl)
{
	CHECK_PARAM_NOT_NULL(sl);

	if (!svga_is_open)
	  error_handler ("attempt to set svga Style before svga is opened");

	if (debug_mode && logfile) {
		fprintf (logfile,"In SVGA: setting the style : ");
		style_dump (logfile, sl);
		fprintf (logfile,"\n");
	}

	style_fix_font (svga_dpy, sl);

	if (!sl->fontdata) 
		error_handler ("font data not obtained in set_style");

	style_copy (&sdata.style, sl);

	if (!sdata.style.fontdata) 
		error_handler ("font data not obtained in set_style");
	
	sdata.fg = color_table_lookup_by_rgb (sl->fg);
	sdata.bg = color_table_lookup_by_rgb (sl->bg);

	if (sdata.fg < 0 || sdata.bg < 0)
		error_handler ("unable to assign color");
}


static Dimension
svga_get_font_height ()
{
	if (!sdata.style.fontdata) 
		return 0;
	else
		return 
		  sdata.style.fontdata->ascent + sdata.style.fontdata->descent;
}


static int
svga_set_font_complete (String family, FontSize size, unsigned short attr) {
	CHECK_PARAM_NOT_NULL(family);

	if (!svga_is_open)
	  error_handler ("attempt to set svga character font before svga is opened");

	if (debug_mode && logfile)
	  fprintf (logfile, "set font complete called, family=%s, size=%d\n", family?(char*)family:"none",(int)size);

	if (!family) family = svga_dpy->family;

	/* Whether we have font info or not, update the
	 * style.
	 */
	sdata.style.family = my_strdup (family);
	sdata.style.size = size;
	sdata.style.flags = attr;
	sdata.style.fontdata = NULL;
	style_fix_font (svga_dpy, &sdata.style);

	if (debug_mode && logfile) {
		fprintf (logfile, "style changed to: ");
		style_dump (logfile, &sdata.style);
		fprintf (logfile, "\n");
	}
	return TRUE;
}


static int
svga_reset_font_complete () {
	if (!svga_is_open)
	  error_handler ("attempt to set svga character font before svga is opened");

	return svga_set_font_complete (svga_dpy->family,svga_dpy->fontsize,0);
}


static int 
svga_set_font (String family) 
{
	if (!family) family = svga_dpy->family;

	return svga_set_font_complete (family, sdata.style.size, sdata.style.flags);
}


static int
svga_set_size (FontSize size) {
	return svga_set_font_complete (sdata.style.family, size, sdata.style.flags);
}


static int
svga_set_attr (unsigned short attr) {
	return svga_set_font_complete (sdata.style.family, sdata.style.size, attr);
}



static void
svga_set_callback (int which, int (*func)(DisplayDevice*,Coordinate,Coordinate,Coordinate,Coordinate)) 
{
	if (!svga_is_open)
	  error_handler ("attempt to set svga callback before svga is opened");

	switch (which) {
	case DISPLAY_REFRESH_CALLBACK:
		svga_dpy->refresh_callback = func;
		break;
	case DISPLAY_RESIZE_CALLBACK:
		svga_dpy->resize_callback = func;
		break;
	default:
		error_handler ("invalid callback identifier");
	}
}


/* Very slow routine to scroll up using SVGALIB */
void
svga_scroll_up (Dimension amount)
{
	int i;
	unsigned char *buffer;
	int wid, ht;
	wid = svga_dpy->width;
	ht = svga_dpy->height;

	svga_pointer_hide();

#ifdef TURBO_C
	/* First try to allocate the entire space. If that works,
	 * the scrolling will be faster.
	 */
	buffer = (char*) my_malloc (wid*amount+10);
	if (buffer) {
		getimage(0,i,wid-1,i+amount-1,buffer);
/* XX - ???? */	putimage(0,i-amount,buffer,COPY_PUT);
	} else {
		buffer = (char*) my_malloc (wid);
		if (!buffer) error_handler ("cannot allocate scroll buffer");

		for (i=amount; i<ht; i++) {
			getimage(0,i,wid-1,i,buffer);
			putimage(0,i-amount,buffer,COPY_PUT);
		}
	}
#else
	buffer = (char*) my_malloc (wid);
	if (!buffer) error_handler ("cannot allocate scroll buffer");

	for (i=amount; i<ht; i++) {
		vga_getscansegment(buffer,0,i,wid);
		vga_drawscansegment(buffer,0,i-amount,wid);
	}
#endif

	my_free(buffer);
	svga_pointer_unhide();
}

/* Very slow routine to scroll down using SVGALIB */
void
svga_scroll_down (Dimension amount)
{
	int i;
	unsigned char *buffer;
	int wid, ht;

	svga_pointer_hide();

	wid = svga_dpy->width;
	ht = svga_dpy->height;
	buffer = (char*) my_malloc (wid);
	if (!buffer) error_handler ("cannot allocate scroll buffer");
	for (i=ht-amount-1; i>=0; i--) {
#ifdef TURBO_C
		getimage (0,i,wid-1,i,buffer);
		putimage (0,i+amount,buffer,COPY_PUT);
#else
		vga_getscansegment(buffer,0,i,wid);
		vga_drawscansegment(buffer,0,i+amount,wid);
#endif
	}

	my_free(buffer);
	svga_pointer_unhide();
}

void
svga_scroll_left (Dimension amount)
{
}

void
svga_scroll_right (Dimension amount)
{
}


/* XX - when background image is working, this will have to copy the image
 * onto the screen.
 */
void
svga_restore_bg (Coordinate ystart, Coordinate yend)
{
#if 0
	unsigned char *buffer;
	int val;
#endif
	int i;
	int wid;
	unsigned char do_hide;

	if (ystart>yend) {
		int tmp=ystart;
		ystart=yend;
		yend=tmp;
	}

	if (ystart<0) ystart=0;
	if (yend>=svga_dpy->height) yend=svga_dpy->height-1;

	do_hide = sdata.pointer_y >= ystart && 
		(sdata.pointer_y+sdata.pointer_height) <= yend;
	if (do_hide)
		svga_pointer_hide();

	wid = svga_dpy->width;

#if 0
	buffer = (unsigned char*) my_malloc (wid);
	if (!buffer) error_handler ("unable to allocate buffer to restore display background");

	val = sdata.default_bg;

	for (i=0; i<wid; i++)
		buffer[i] = val (from image)...

	drawscansegment....

	my_free(buffer);
#endif

	i = _getcolor();
	_setcolor(sdata.bg);
	_fillarea (0,ystart,wid-1,yend);
	_setcolor (i);

	if (do_hide)
		svga_pointer_unhide();
}



Font*
svga_get_font_metrics (String family, FontSize size, unsigned short attr)
{
	Font *f;

	/* If no family or size was specified, that's OK. The parser 
	 * and executor are not required to be aware of fonts beyond
	 * what the HTML file specifies. The user can specify a default
	 * font and size on the command line (XX future feature).
	 */
	if (!family) 
		family=svga_dpy->family;
	if (!size || size==INVALID_FONTSIZE) 
		size=svga_dpy->fontsize;

	f = fonthash_lookup (svga_dpy, family,size,attr);

	return f;
}



DisplayDevice*
svga_init (Dimension max_width, 
	Dimension max_height, 
	Color fg, 
	Color bg,
	Color pointer_fg, 
	Color pointer_border_color,
	String family)
{
	DisplayDevice *dpy;

	CHECK_PARAM_NOT_NULL (family);

	dpy = display_new();
	if(!dpy)
		error_handler ("cannot allocate DisplayDevice object");

	memset ((char*) &sdata, 0, sizeof(sdata));

	fonthash_init ();

	svga_dpy = dpy;

	/* Check the validity of the default font and fontsize.
	 */
	dpy->family = my_strdup(family);
	dpy->fontsize = 12;

	{ 
		Font *f = fonthash_lookup (dpy, family, dpy->fontsize, 0);
		if (!f) return NULL;
	}

	/* Request our display mode (need to try several!)
	 */

#ifndef TURBO_C
	// vga_setmode (G640x480x2);
	dpy->actual_width=        640;
	dpy->actual_height=       480;

	vga_setmode (G640x480x16);

	color_table_init (16);

	/* Set up some predefined pixel values
	 */
	dpy->fg = color_table_lookup_by_rgb (fg);
	dpy->bg = color_table_lookup_by_rgb (bg);
	dpy->white = color_table_lookup ("white");
	dpy->black = color_table_lookup ("black");
	dpy->red = color_table_lookup ("red");
	dpy->green = color_table_lookup ("green");
	dpy->blue = color_table_lookup ("blue");
	dpy->grey1 = color_table_lookup ("blue");
	dpy->grey2 = color_table_lookup ("blue");
	dpy->grey3 = color_table_lookup ("blue");
	dpy->grey4 = color_table_lookup ("blue");
	dpy->grey5 = color_table_lookup ("blue");
	dpy->grey6 = color_table_lookup ("blue");
	dpy->grey7 = color_table_lookup ("blue");

	vga_screenon ();
	{
		int i = vga_getmousetype ();
		fprintf (logfile,"SVGALIB says: mouse type = %d\n", i);
		mouse_init("/dev/mouse", i, MOUSE_DEFAULTSAMPLERATE);
		mouse_setxrange (0, 639);
		mouse_setyrange (0, 479);
		mouse_setposition (320,240);
		mouse_setscale (1);
		mouse_setwrap (0);
		sdata.pointer_x = 320;
		sdata.pointer_y = 240;
		sdata.pointer_width = left_ptr_width;
		sdata.pointer_height = left_ptr_height;
		sdata.pointer_is_visible = FALSE;
		sdata.pointer_is_hidden = FALSE;
	}
#else
{
	int driver = DETECT;
	int mode;
	initgraph (&driver,&mode,"");
	if (graphresult() != grOk) {
		warning_handler ("cannot initialize VGA screen");
	} 

	/* dpy->max_colors = getmaxcolor() + 1; */
	dpy->actual_width = getmaxx();
	dpy->actual_height = getmaxy();
}
#endif

	/* Set up the default clip */
	_setclip (0,0,dpy->actual_width-1, dpy->actual_height-1);

	/* HTML-specific dimensions which are display-specific
	 */
	dpy->heading_space = 8;
	dpy->interparagraph_space = 8;
	dpy->horizontal_line_space = 8;
	dpy->list_indentation_space = 24;
	dpy->interline_padding = 1;
	dpy->table_cell_spacing = 3;
	dpy->table_cell_padding = 0;
	dpy->table_border_size = 0;

	/* Set up scrollbars (enabling both)
	 */
	dpy->vscroll_width = VSCROLL_WIDTH;
	dpy->hscroll_height = HSCROLL_HEIGHT;

	/* Drawable area 
	 */
	dpy->width = max_width - dpy->vscroll_width;
	dpy->height = max_height - dpy->hscroll_height;
	
	/* Set up methods.
	 */
	dpy->refresh =     &svga_refresh;
	dpy->resize =      &svga_resize;
	dpy->clear =       &svga_clear;
	dpy->get_font_metrics = &svga_get_font_metrics;
	dpy->set_font_complete = &svga_set_font_complete;
	dpy->reset_font_complete = &svga_reset_font_complete;
	dpy->set_style=    &svga_set_style;
	dpy->get_font_height=    &svga_get_font_height;
	dpy->set_font=     &svga_set_font;
	dpy->set_size=     &svga_set_size;
	dpy->set_attr=     &svga_set_attr;

	dpy->set_colors =  &svga_set_colors;
	dpy->reset_colors =  &svga_reset_colors;
	dpy->rgb_to_pixel= &color_table_lookup_by_rgb;
	dpy->color_to_pixel= &color_table_lookup;

	dpy->set_clip =    &svga_set_clip;
	dpy->reset_clip =  &svga_reset_clip;

	dpy->draw_text =   &svga_draw_text;
	dpy->draw_rectangle=     &svga_draw_rectangle;
	dpy->draw_3d_rectangle=  &svga_draw_3d_rectangle;
	dpy->filled_rectangle=   &svga_filled_rectangle;
	dpy->highlight_area =   &svga_highlight_area;
	dpy->unhighlight_area =   &svga_unhighlight_area;
	dpy->draw_x =      &svga_draw_x;
	dpy->draw_hline =  &svga_draw_hline;
	dpy->draw_vline =  &svga_draw_vline;
	dpy->draw_line =   &svga_draw_line;
	dpy->text_dims =   &svga_text_dims;
	dpy->get_event =   &svga_get_event;
	dpy->close =       &svga_close;
	dpy->dump =        &svga_dump;
	dpy->scroll_up =   &svga_scroll_up;
	dpy->scroll_down = &svga_scroll_down;
	dpy->scroll_left = &svga_scroll_left;
	dpy->scroll_right= &svga_scroll_right;
	dpy->vscroll_enable= &svga_vscroll_enable;
	dpy->vscroll_disable= &svga_vscroll_disable;
	dpy->hscroll_enable= &svga_hscroll_enable;
	dpy->hscroll_disable= &svga_hscroll_disable;
	dpy->vscroll_move= &svga_vscroll_move;
	dpy->hscroll_move= &svga_hscroll_move;
	dpy->restore_background= &svga_restore_bg;
	dpy->allocate_buffers = &svga_allocate_buffers;
	dpy->set_callback= &svga_set_callback;
	dpy->convert_special_expr=   &svga_special_expr;

	dpy->pointer_enable = &svga_pointer_enable;
	dpy->pointer_disable = &svga_pointer_disable;
	dpy->pointer_hide = &svga_pointer_hide;
	dpy->pointer_unhide = &svga_pointer_unhide;

	dpy->get_image = &svga_get_image;
	dpy->put_image = &svga_put_image;

	dpy->draw_twoline_centered_popup = &svga_draw_twoline_centered_popup;

	style_reset (&sdata.style);

	/* Set up default colors.
	 */
	dpy->fgcolor = fg;
	dpy->bgcolor = bg;
	dpy->pointer_fgcolor = pointer_fg;
	dpy->pointer_bordercolor = pointer_border_color;
	dpy->highlight_color = 0x00ffff;

	sdata.fg = color_table_lookup_by_rgb (fg);
	sdata.bg = color_table_lookup_by_rgb (bg);
	sdata.style.fg = dpy->fgcolor;
	sdata.style.bg = dpy->bgcolor;
	sdata.pointer_fg = color_table_lookup_by_rgb (pointer_fg);
	sdata.pointer_border_color = color_table_lookup_by_rgb (pointer_border_color);
	sdata.highlight_bg = color_table_lookup_by_rgb (dpy->highlight_color);

	if (sdata.fg < 0 || sdata.bg < 0)
		error_handler ("unable to assign color");

#if 0
	/* color test */
	{
		int i,j;
		static unsigned char scanline [1000];
		unsigned char *s=scanline;
		j=0;
		for (i=0;i<640;i++) {
			*s = i/40;
			++s;
		}
		for (i=0;i<240; i++) {
			vga_drawscanline (i,scanline);
		}
	}
#endif

	svga_is_open = TRUE;

	svga_reset_font_complete ();

	_setcolor(sdata.bg);
	_fillarea(0,0,dpy->width-1,dpy->height-1);
	_setcolor(sdata.fg);

	svga_hscroll_draw ();
	svga_vscroll_draw ();

	return dpy;
}


