
/*=============================================================================
   beest, an HTML viewer.
   Copyright (C) 2000 Zachary Thayer Smith

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   The author is reachable by electronic mail at zeyr@excite.com.
=============================================================================*/

/*----------------------------------------------------------------------
 * Module name:    word.c
 * Author name:    Zach Smith
 * Create date:    1 Oct 2000
 * Last change:    26 Nov 2000
 * Purpose:        Word class, equivalent to a text word or a tag;
 *                 Segment class for breaking apart tags.
 *----------------------------------------------------------------------
 * Changes:
 *--------------------------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>


#include "defs.h"
#include "malloc.h"
#include "display.h"
#include "main.h"
#include "word.h"
#include "line.h"
#include "style.h"
#include "special.h"
#include "group.h"
#include "parse.h"
#include "line.h"
#include "error.h"
#include "table.h"



/*-----------------------Word Class Functionality-----------------------*/

Word* 
word_new (String s) {
	Word* w;
	CHECK_PARAM_NOT_NULL(s);
	w=(Word*) my_malloc(sizeof(Word));
	if(!w) error_handler ("cannot allocate Word");
	memset (w,0,sizeof(Word));
	w->str=my_strdup(s);
	w->x = w->y = 0;
#if 0
	/* These happen automatically during memset().
	 */
	w->next=w->child=w->next_in_line = NULL;
	w->flags = 0;
	w->width = 0;
	w->ascent = 0;
	w->descent = 0;
	w->tag.datum = 0;
#endif
	return w;
}

void 
word_free (Word *w) {
	CHECK_PARAM_NOT_NULL(w);
	if (!w->str)
		error_handler ("Word object lacks string");
	free (w->str);
	free (w);

	/* need to free segments XX */
}


void
word_apply_offset (Word *w, Coordinate off_x, Coordinate off_y) 
{
	return;
#if 0
	while (w) {
		if (w->child) 
			word_apply_offset (w->child, off_x, off_y);

		w->x += off_x;
		w->y += off_y;
/* printf ("APPLY moved '%s' to %d,%d\n", w->str,w->x,w->y); */

		w=w->next;
	}
#endif
}


Segment*
segment_new (String str)
{
	Segment *seg;
	CHECK_PARAM_NOT_NULL(str);
	seg = (Segment*)  my_malloc (sizeof (Segment));
	if (!seg) error_handler ("unable to allocate Segment object");

	seg->str = my_strdup(str);
	seg->next=NULL;
	return seg;
}

void
word_append_segment (Word *w, Segment *s)
{
	CHECK_PARAM_NOT_NULL(w);
	CHECK_PARAM_NOT_NULL(s);
	if(!w->segments) {
		w->segments = s;
		w->last_segment = s;
		w->total_segments = 1;
	} else {
		w->last_segment->next = s;
		w->last_segment = s;
		s->next=NULL;
		w->total_segments++;
	}
}


/* Convert a tag into several segments, e.g.
 *   <A HREF=FOO.HTML>
 * becomes
 *   A
 *   HREF
 *   =
 *   FOO.HTML
 */
int
word_divide_tag (Word *w) 
{
#define MAX_SEGMENT_LENGTH (256)
	char buffer [MAX_SEGMENT_LENGTH];
	int ch;
	unsigned short ix, in_str, quotechar;
	String s;

	CHECK_PARAM_NOT_NULL(w);

	/* No string? */
	s = w->str;
	if (!s) return FALSE;

	/* Not a tag? */
	if (*s!='<') return FALSE;

	/* Skip < */
	s++;

	/* Skip whitespace */
	if (*s==' ') {
		while(*s==' ') s++;
	}

	/* Make them segments, make 'em good. */
	while ((ch = *s++)) {

		/* End o' tag? */
		if (ch=='>') {
			break;
		}

		/* Equals is always 1 segment */
		if (ch=='=') {
			word_append_segment(w,segment_new("="));
		}
		else
		{
			ix=0;
			in_str=FALSE;
			quotechar=0;
			do {
				if (in_str && ch==quotechar) break;
				if (!in_str) {
					if (ch==' ') break;
					else if (ch=='>') { s--; break; }
					else if (ch=='=') { s--; break; }
					else if (ch=='"' || ch=='\'') {
						in_str=TRUE;
						quotechar=ch; 
						ch = *s++;
						continue;
					}
				}
				buffer[ix++] = ch;
				if (ix==(MAX_SEGMENT_LENGTH-1)) 
					break;
				ch = *s++;
			}
			  while(ch);

			buffer[ix]=0;
			word_append_segment(w,segment_new(buffer));
		}

		/* Skip whitespace */
		if (*s==' ') {
			while(*s==' ') s++;
		}
	}

	/* Generate the tag number */
	w->tag_number = taghash_lookup_number (w->segments->str);
	return TRUE;
}


String
word_fetch_tag_var_value (Word *w, String varname)
{
	Segment *s;
	Segment *s2;

	CHECK_PARAM_NOT_NULL(w);
	CHECK_PARAM_NOT_NULL(varname);

	s = w->segments;

	/* Skip the tag itself */
	if (s) s=s->next;

	while(s) {
		if (!strcmp (varname, s->str)) {
			s2 = s->next;

			/* If the var has no "=" following it,
			 * assume it's a flag.
			 */
			if (!s2) 
				return "PRESENT";

			if ('=' != s2->str[0])
				return "PRESENT";

			s2 = s2->next;

			if (!s2) return NULL;

			return s2->str;
		}

		s=s->next;
	}
	return NULL;
}


